import os
import click

from .registry import service, package_language, repository_host


@click.group()
def main():
    pass

def ensure_config_run():
    if os.path.exists('config.py'):
        with open('config.py', 'r') as f:
            exec(f.read())


def ask_questions(name, handler):

    function = handler['function']
    arguments = handler['arguments']

    parameters = []
    for argument in arguments:
        # TODO: recongize default values from functions
        value = click.prompt(argument['description'], type=argument['type'], hide_input='password' in argument['name'])
        parameters.append(value)

    print('Running setup for {0}... '.format(name), end='')
    try:
        return function(*parameters)
    finally:
        click.echo(click.style('done', 'green'))


@click.command()
def quickstart():

    ensure_config_run()

    click.echo(click.style("Welcome to LetThereBe!", 'red'))
    print('')
    print('Please answer the following questions to set up your project (just press Enter to accept default values when available)')

    # Set up repo

    print('')
    click.echo(click.style("Step 1 - Setting up repository", 'magenta'))
    print('')

    available_repo_hosts = sorted(repository_host.members)

    if 'github' in available_repo_hosts:
        default_repo_host = 'github'
    else:
        default_repo_host = available_repo_hosts[0]

    repo_host = click.prompt('Which service would you like to set up your repository on?',
                             default=default_repo_host, type=click.Choice(available_repo_hosts))

    repo = ask_questions(repo_host, repository_host.members[repo_host])

    # Set up language

    print('')
    click.echo(click.style("Step 2 - Setting up the package layout", 'magenta'))
    print('')

    available_languages = sorted(package_language.members)

    if 'python' in available_languages:
        default_language = 'python'
    else:
        default_language = available_languages[0]

    language = click.prompt('Which language would you like to set up a project for?',
                            default=default_language, type=click.Choice(available_languages))

    local_path = ask_questions(language, package_language.members[language])

    # Add code to repository
    repo.add_directory_to_repo(local_path, '.', 'Initial commit with package layout')

    # Set up services

    print('')
    click.echo(click.style("Step 3 - Setting up services", 'magenta'))
    print('')

    available_services = sorted(service.members)

    services = click.prompt('Which services do you want to set up? (comma separated)', default=', '.join(available_services))

    for s in services.split(','):
        ask_questions(s.strip(), service.members[s.strip()])

    # Give final instructions_to_get_repo
    print('')
    click.echo(click.style("Epilogue", 'magenta'))
    print('')
    print(repo.instructions_to_get_repo())

main.add_command(quickstart)



from cookiecutter.main import cookiecutter
import json
import urllib.request
import tempfile

def projskeleton(infodict):
    gitlocation = {
        "C++": "https://github.com/13coders/cookiecutter-kata-gtest", 
        "Python": "https://github.com/audreyr/cookiecutter-pypackage"
    }


    txt = json.dumps(infodict)
    with open("cookiecutter.json", "w") as f:
        f.write(json.dumps(txt, ensure_ascii=False))

    tmpd =tempfile.mkdtemp()
    cookiecutter(gitlocation[infodict['language_name']], 
                 extra_context=infodict,
                 config_file="cookiecutter.json",
                 overwrite_if_exists=True, 
                 default_config=False,
                 no_input=True, 
                 output_dir=tmpd)

    return(tmpd + '/' + infodict['project_name'])



import os

from .registry import repository_host


class Repository(object):
    """
    Unified object to represent repositories
    """

    def add_directory_to_repo(local_path, remote_path, message):
        raise NotImplementedError()

    def instructions_to_get_repo(self):
        raise NotImplementedError()


GITHUB_EPILOGUE = """
Your repository has been set up! You can see it in all its glory at:

    {html_url}

If you want to get it locally, you will need to make sure you have git installed,
then you will be able to clone it with:

    git clone {clone_url}

Enjoy!
"""


class GitHubRepository(Repository):

    def __init__(self, repo):
        self._repo = repo

    def add_directory_to_repo(self, local_path, remote_path, message):
        """
        Add a new file to a new branch in a GitHub repo
        """

        from github import InputGitTreeElement

        commit = self._repo.get_branch('master').commit
        git_commit = self._repo.get_git_commit(commit.sha)

        tree_input = []

        # List files
        for root, dirnames, filenames in os.walk(local_path):
            for filename in filenames:

                full_path = os.path.join(root, filename)
                repo_path = os.path.relpath(full_path, local_path)

                with open(full_path, 'r') as f:
                    content = f.read()

                blob = self._repo.create_git_blob(content=content, encoding='utf-8')
                tree_input.append(InputGitTreeElement(os.path.normpath(os.path.join(remote_path, repo_path)),
                                                      "100644", "blob", sha=blob.sha))

        # We make a new tree, commit, and branch
        tree = self._repo.create_git_tree(tree=tree_input)
        commit = self._repo.create_git_commit(tree=tree, message=message, parents=[git_commit])
        master = self._repo.get_git_ref('heads/master')
        master.edit(commit.sha)

    def instructions_to_get_repo(self):
        return GITHUB_EPILOGUE.strip().format(html_url=self._repo.html_url,
                                              clone_url=self._repo.clone_url)


@repository_host('github')
def set_up_github_repo(login_or_token, password, name, description):
    """
    Set up an empty GitHub repository

    Parameters
    ----------
    login_or_token : str
        What is your GitHub username?
    password : str
        what is your GitHub password?
    name : str
        What is the repository name?
    description : str
        Can you provide a one-line description for the package?
    """
    import github
    gh = github.Github(login_or_token=login_or_token,
                       password=password)
    user = gh.get_user()
    repo = user.create_repo(name, description=description, auto_init=True)
    return GitHubRepository(repo)



from numpydoc.docscrape import NumpyDocString


def extract_arguments(function):
    """
    Given a Python function, parse the numpydoc docstring and return a standard
    list of arguments.
    """
    doc = NumpyDocString(function.__doc__)
    arguments = []
    for arg_name, arg_type, arg_description in doc['Parameters']:
        argument = {}
        argument['name'] = arg_name
        argument['type'] = type(arg_type.split(',')[0].strip())
        argument['optional'] = 'optional' in arg_type
        argument['description'] = arg_description[0]
        arguments.append(argument)
    return arguments



def set_up_python_package(name, tests=True):
    """
    Set up a basic Python package

    Parameters
    ----------
    name : str
        What is the package name?
    tests : bool, optional
        Should tests be included?
    """
    pass

#
# @language('python', help={'name': 'Specify the name of your package',
#                           'tests': 'Should example tests be included?'})
# def set_up_python_package(name, tests=True):
#     pass



from .utils import extract_arguments


class HandlerRegistry(object):
    """
    Base container class to hold groups of handlers
    """

    def __init__(self):
        self.members = {}

    def __call__(self, identifier):
        """
        This is provided so that registry instances can be used
        as decorators.
        """

        def wrapper(function):
            arguments = extract_arguments(function)
            self.members[identifier] = {'function': function, 'arguments': arguments}
        return wrapper


repository_host = HandlerRegistry()
package_language = HandlerRegistry()
service = HandlerRegistry()



from setuptools import setup

from lettherebe import __version__

entry_points = """
[console_scripts]
lettherebe = lettherebe.main:main
"""

setup(
    version=__version__,
    url="https://github.com/LetThereBe/lettherebe",
    name="lettherebe",
    description='LetThereBe Python library and command-line scripts',
    packages=['lettherebe'],
    install_requires=['requests', 'pygithub', 'click', 'colorama', 'cookiecuttter'],
    license='BSD',
    author="Mayeul d'Avezac, Ilektra Christidi, Alice Harpole, David Perez Suarez, Thomas Robitaille, Sinan Shi",
    author_email="thomas.robitaille@gmail.com",
    classifiers=[
        'Development Status :: 1 - Planning',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: BSD License',
    ],
    entry_points=entry_points
)



from .registry import service


@service('readthedocs')
def set_up_readthedocs():
    """
    Set up readthedocs for the repository

    Parameters
    ----------
    """
    pass


@service('travis')
def set_up_travis():
    """
    Set up Travis for the repository

    Parameters
    ----------
    """
    pass



from .registry import package_language

from cookiecutter.main import cookiecutter
import json
import urllib.request
from lettherebe import ccwrapper

def downloadtemplate(url):
    response = urllib.request.urlopen(url)
    data = response.read()      
    text = data.decode('utf-8')
    return(text)


@package_language('python')
def set_up_python_package(full_name, email, github_username, 
                          project_name, project_short_description, 
                          license, version = '0.1.0'):
    """
    Set up a basic Python package

    Parameters
    ----------
    full_name : str
        Author's full name
    email : str
        Author's email address
    github_username : str
        Author's GitHub username
    project_name : str
        The project name
    project_short_description : str
        Give a short description of your project
    license : {'MIT', 'BSD', 'ISC', 'Apache', 'GNU', 'Not open'}
        Choose a licence for your project
    version : str, optional
        First version of your package. Default is 0.1.0
    """

    jsonlocation = "https://raw.githubusercontent.com/audreyr/cookiecutter-pypackage/master/cookiecutter.json"
    jsoncontent = downloadtemplate(jsonlocation)
    jsondict = json.loads(jsoncontent)

    licensedict = {"MIT": "MIT license",
                   "BSD": "BSD license",
                   "ISC": "ISC license",
                   "Apache": "Apache Software License 2.0",
                   "GNU": "GNU General Public License v3",
                   "Not open": "Not open source"}
    
    jsondict['full_name'] = full_name 
    jsondict['email'] = email
    jsondict['github_username'] = github_username
    jsondict['project_name'] = project_name
    jsondict['project_short_description'] = project_short_description
    jsondict['open_source_license'] = licensedict[license]
    jsondict['version'] = version
    jsondict['use_pytest'] = "y"
    jsondict['language_name'] = "Python"

    return(ccwrapper.projskeleton(jsondict))
    pass


@package_language('cpp')
def set_up_cpp_package(project_name, license):
    """
    Set up a basic C++ package with gTest infrastructure

    Parameters
    ----------
    project_name : str
        The project name
    license : {'MIT', 'Apache', 'GNU'}
        Choose a licence for your project
    """

    jsonlocation = "https://raw.githubusercontent.com/13coders/cookiecutter-kata-cpputest/master/cookiecutter.json"
    jsoncontent = downloadtemplate(jsonlocation)
    jsondict = json.loads(jsoncontent)

    licensedict = {"MIT": "MIT license",
                   "Apache": "Apache Software License 2.0",
                   "GNU": "GNU General Public License v3"}
    
    jsondict['kata'] = project_name
    jsondict['project_name'] = project_name
    jsondict['license'] = licensedict[license]
    jsondict['language_name'] = "C++"
 
    return(ccwrapper.projskeleton(jsondict))
    pass
